#
# Sonar Scanner Python
# Copyright (C) 2011-2024 SonarSource SA.
# mailto:info AT sonarsource DOT com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful,
#
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
import functools
import logging
from logging import Logger
from typing import Optional


class ApplicationLogger:
    _log: Optional[Logger] = None

    @classmethod
    def get_logger(cls) -> Logger:
        if not cls._log:
            cls._log = logging.getLogger("main")
            cls._setup_logger(cls._log)
        return cls._log

    @staticmethod
    def _setup_logger(log: Logger):
        log.setLevel(logging.INFO)
        handler = logging.StreamHandler()
        log.addHandler(handler)

    @classmethod
    def set_debug(cls, debug: bool) -> None:
        if debug:
            cls.get_logger().setLevel(logging.DEBUG)
            cls.get_logger().exception = functools.partial(cls.get_logger().exception, exc_info=True)
        else:
            cls.get_logger().setLevel(logging.INFO)
            cls.get_logger().exception = functools.partial(cls.get_logger().exception, exc_info=False)



#
# Sonar Scanner Python
# Copyright (C) 2011-2024 SonarSource SA.
# mailto:info AT sonarsource DOT com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful,
#
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
from __future__ import annotations
import os
import platform
import pyfiglet
import shutil
import urllib.request
from urllib.error import HTTPError

from pysonar_scanner.logger import ApplicationLogger
from pysonar_scanner.configuration import Configuration
from pysonar_scanner.utils.binaries_utils import write_binaries, unzip_binaries
from pysonar_scanner.scanner import Scanner

systems = {"Darwin": "macosx", "Windows": "windows"}


class Environment:
    cfg: Configuration

    # a full download path for a scanner has the following shape:
    # https://binaries.sonarsource.com/Distribution/sonar-scanner-cli/sonar-scanner-cli-${version}-${os}.zip
    scanner_base_url: str = "https://binaries.sonarsource.com/Distribution/sonar-scanner-cli/sonar-scanner-cli"

    def __init__(self, cfg: Configuration):
        self.cfg = cfg
        self.log = ApplicationLogger.get_logger()

    def scan(self):
        try:
            self.setup()
            self.scanner().scan()
        finally:
            self.cleanup()

    def setup(self):
        self.cleanup()
        if self._is_sonar_scanner_on_path():
            self.cfg.sonar_scanner_executable_path = "sonar-scanner"
        else:
            system_name = systems.get(platform.uname().system, "linux")
            self._install_scanner(system_name)
            sonar_scanner_home = os.path.join(
                self.cfg.sonar_scanner_path,
                f"sonar-scanner-{self.cfg.sonar_scanner_version}-{system_name}",
            )
            self.cfg.sonar_scanner_executable_path = os.path.join(sonar_scanner_home, "bin", "sonar-scanner")

        ascii_banner = pyfiglet.figlet_format("Sonar Scanner")
        self.log.info(ascii_banner)

    def scanner(self) -> Scanner:
        return Scanner(self.cfg)

    def _is_sonar_scanner_on_path(self) -> bool:
        return shutil.which("sonar-scanner") is not None

    def cleanup(self):
        if os.path.exists(self.cfg.sonar_scanner_path):
            shutil.rmtree(self.cfg.sonar_scanner_path)

    def _install_scanner(self, system_name: str):
        os.mkdir(self.cfg.sonar_scanner_path)
        # Download the binaries and unzip them
        scanner_zip_path = self._download_scanner_binaries(
            self.cfg.sonar_scanner_path, self.cfg.sonar_scanner_version, system_name
        )
        unzip_binaries(scanner_zip_path, self.cfg.sonar_scanner_path)
        os.remove(scanner_zip_path)
        self._change_permissions_recursive(self.cfg.sonar_scanner_path, 0o777)

    def _change_permissions_recursive(self, path, mode):
        for root, dirs, files in os.walk(path, topdown=False):
            for dir in [os.path.join(root, d) for d in dirs]:
                os.chmod(dir, mode)
            for file in [os.path.join(root, f) for f in files]:
                os.chmod(file, mode)

    def _download_scanner_binaries(self, destination: str, scanner_version: str, system_name: str) -> str:
        try:
            scanner_res = urllib.request.urlopen(f"{self.scanner_base_url}-{scanner_version}-{system_name}.zip")
            scanner_zip_path = os.path.join(destination, "scanner.zip")
            write_binaries(scanner_res, scanner_zip_path)
            return scanner_zip_path
        except HTTPError as error:
            self.log.error(f"ERROR: could not download scanner binaries - {error.code} - {error.msg}")
            raise error



#
# Sonar Scanner Python
# Copyright (C) 2011-2024 SonarSource SA.
# mailto:info AT sonarsource DOT com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful,
#
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
from __future__ import annotations
from logging import Logger
from threading import Thread
from subprocess import Popen, PIPE

from pysonar_scanner.logger import ApplicationLogger
from pysonar_scanner.configuration import Configuration


class Scanner:
    cfg: Configuration
    log: Logger

    def __init__(self, cfg: Configuration):
        self.cfg = cfg
        self.log = ApplicationLogger.get_logger()

    def scan(self):
        process = self.execute_command()
        output_thread = Thread(target=self._log_output, args=(process.stdout,))
        error_thread = Thread(target=self._log_output, args=(process.stderr,))
        return self.process_output(output_thread, error_thread, process)

    def process_output(self, output_thread: Thread, error_thread: Thread, process: Popen) -> int:
        output_thread.start()
        error_thread.start()
        process.wait()
        output_thread.join()
        error_thread.join()

        return process.returncode

    def execute_command(self) -> Popen:
        cmd = self.compute_command()
        return Popen(cmd, stdout=PIPE, stderr=PIPE)

    def compute_command(self) -> list[str]:
        if not self.cfg.sonar_scanner_executable_path:
            raise ValueError("No executable path provided")
        return [self.cfg.sonar_scanner_executable_path] + self.cfg.scan_arguments

    def _log_output(self, stream: list[bytes]):
        for line in stream:
            decoded_line = line.decode("utf-8").rstrip()
            self.log.info(decoded_line)



#
# Sonar Scanner Python
# Copyright (C) 2011-2024 SonarSource SA.
# mailto:info AT sonarsource DOT com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful,
#
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
from pysonar_scanner.configuration import Configuration
from pysonar_scanner.environment import Environment
from pysonar_scanner.logger import ApplicationLogger


def scan():
    log = ApplicationLogger.get_logger()
    cfg = Configuration()
    try:
        cfg.setup()
        env = Environment(cfg)
        env.scan()
    except Exception as e:
        log.exception("Error during SonarScanner execution: %s", str(e))
        if not cfg.is_debug():
            log.info("Re-run SonarScanner using the -X switch to enable full debug logging.")


if __name__ == "__main__":
    scan()



#
# Sonar Scanner Python
# Copyright (C) 2011-2024 SonarSource SA.
# mailto:info AT sonarsource DOT com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful,
#
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
from __future__ import annotations

import argparse
import os
import sys
from logging import Logger
from typing import Union

import toml

from pysonar_scanner.logger import ApplicationLogger


class Configuration:
    log: Logger
    sonar_scanner_executable_path: str
    sonar_scanner_path: str
    sonar_scanner_version: str
    scan_arguments: list[str]
    wrapper_arguments: argparse.Namespace

    def __init__(self):
        self.log = ApplicationLogger.get_logger()
        self.sonar_scanner_path = ".scanner"
        self.sonar_scanner_version = "6.0.0.4432"
        self.sonar_scanner_executable_path = ""
        self.scan_arguments = []
        self.wrapper_arguments = argparse.Namespace(debug=False, read_project_config=False)
        self.unknown_arguments = []

    def setup(self) -> None:
        """This is executed when run from the command line"""
        self._read_wrapper_arguments()
        ApplicationLogger.set_debug(self.is_debug())
        self.scan_arguments = self._read_toml_args()
        # The sonar-scanner-cli may crash with an error when receiving unexpected arguments
        # Therefore, arguments only expected by the sonar-scanner-python are filtered out.
        self._append_common_arguments()
        # If duplicate properties are provided, newer values will override older ones.
        # We therefore read CLI arguments last so that they have priority over toml configuration.
        self.scan_arguments.extend(self.unknown_arguments)

    def _append_common_arguments(self):
        if self.wrapper_arguments.debug:
            self.scan_arguments.append("-X")
        if self.wrapper_arguments.project_home is not None:
            self.scan_arguments.append(f"-Dproject.home={self.wrapper_arguments.project_home}")

    def _read_wrapper_arguments(self):
        argument_parser = argparse.ArgumentParser()
        argument_parser.add_argument("-toml.path", "-Dtoml.path", "--toml.path", dest="toml_path")
        argument_parser.add_argument("-project.home", "-Dproject.home", "--project.home", dest="project_home")
        argument_parser.add_argument("-X", action="store_true", dest="debug")
        argument_parser.add_argument(
            "-read.project.config",
            "-Dread.project.config",
            "--read-project-config",
            "-read-project-config",
            action="store_true",
            dest="read_project_config",
        )
        self.wrapper_arguments, self.unknown_arguments = argument_parser.parse_known_args(args=sys.argv[1:])

    def _read_toml_args(self) -> list[str]:
        scan_arguments: list[str] = []
        toml_data: dict = {}
        try:
            toml_data = self._read_toml_file()
        except OSError as e:
            self.log.exception("Error while opening .toml file: %s", str(e))
        if self.wrapper_arguments.read_project_config:
            common_toml_properties = self._extract_common_properties(toml_data)
            for key, value in common_toml_properties.items():
                self._add_parameter_to_scanner_args(scan_arguments, key, value)
        sonar_properties = self._extract_sonar_properties(toml_data)
        for key, value in sonar_properties.items():
            self._add_parameter_to_scanner_args(scan_arguments, key, value)
        return scan_arguments

    def _extract_sonar_properties(self, toml_properties):
        if "tool" not in toml_properties.keys():
            return {}
        tool_data = toml_properties["tool"]
        if not isinstance(tool_data, dict) or "sonar" not in tool_data.keys():
            return {}
        sonar_properties = tool_data["sonar"]
        return sonar_properties if isinstance(sonar_properties, dict) else {}

    def _extract_common_properties(self, toml_properties):
        properties = {}
        if "tool" in toml_properties.keys() and "poetry" in toml_properties["tool"]:
            poetry_properties = toml_properties["tool"]["poetry"]
            properties = self._extract_from_poetry_properties(poetry_properties)
        return properties

    def _extract_from_poetry_properties(self, poetry_properties):
        result = {}
        if "name" in poetry_properties:
            result["project.name"] = poetry_properties["name"]
        if "version" in poetry_properties:
            result["projectVersion"] = poetry_properties["version"]
        # Note: Python version can be extracted from dependencies.python, however it
        # may be specified with constraints, e.g ">3.8", which is not currently supported by sonar-python
        return result

    def _add_parameter_to_scanner_args(self, scan_arguments: list[str], key: str, value: Union[str, dict]):
        if isinstance(value, str):
            scan_arguments.append(f"-Dsonar.{key}={value}")
        if isinstance(value, dict):
            for k, v in value.items():
                self._add_parameter_to_scanner_args(scan_arguments, f"{key}.{k}", v)

    def _read_toml_file(self) -> dict:
        toml_file_path = self._get_toml_file_path()
        if not os.path.isfile(toml_file_path):
            return {}
        with open(toml_file_path, "r") as file:
            # TODO: actually search for pyproject.toml
            toml_data = file.read()
            return toml.loads(toml_data)

    def _get_toml_file_path(self) -> str:
        if self.wrapper_arguments.toml_path is not None:
            return self.wrapper_arguments.toml_path
        elif self.wrapper_arguments.project_home is not None:
            return os.path.join(self.wrapper_arguments.project_home, "pyproject.toml")
        else:
            return os.path.join(os.curdir, "pyproject.toml")

    def is_debug(self) -> bool:
        return self.wrapper_arguments.debug
